#include "msntrans.h"
#include "filetransfer.h"


void mt_msnftp_connect_parse(mio m, int state, void *arg, char *buffer, int bufsz) {
	msnftp_invite invite = (msnftp_invite) arg;
	
	if(invite->stage == INCOMING_BINARYMODE) {
		if(buffer[0] == 1) {
			mt_stream_write(invite->st, "BYE 16777989");
			invite->stage = TRANSFERDONE;
			invite->filedata[j_atoi(invite->applicationfilesize, 0)] = 0;
			log_debug(ZONE, "--- TOTAL FILE DATA --- \n\n%s\n\n ------ END TOTOL FILE DATA ------", invite->filedata);
			return;
		}
//		unsigned int sz=256*(int)buffer[2] + (int)buffer[1];
		strncpy(invite->filedata, buffer, bufsz);
		log_debug(ZONE, "Buffering binary data");
	}
	else {
		log_debug(ZONE, "Falling back to ordinary stream read");
		mt_stream_read(m, state, arg, buffer, bufsz);
	}
}

result mt_msnftp_connect_commands(mpacket mp, void *arg) {
	msnftp_invite invite = (msnftp_invite) arg;
	int i = 0;
	if(mp == NULL) {
		return r_DONE;
	}
	
	log_debug(ZONE, "mt_msnftp_connect_packets with %s", mt_packet_data(mp, 0));
	
	if(invite->stage == INCOMING_TRANSFERSTARTED) {
		if(j_strcmp(mt_packet_data(mp, 0), "VER") == 0) {
			for(i = 1; i < mp->count; i++) {
				if(j_strcmp(mt_packet_data(mp, i), "MSNFTP") == 0) {
					mt_stream_write(invite->st,"USR %s %s\r\n",invite->s->user, invite->authcookie);
					return r_DONE;
				}
				else {
					log_debug(ZONE, "Unknown protocol for file transfer: %s", mt_packet_data(mp, i));
				}
			}
		}
		else if(mp->count == 2 && j_strcmp(mt_packet_data(mp, 0), "FIL") == 0) {
			invite->applicationfilesize = pstrdup(invite->p, mt_packet_data(mp, 1));
			mt_stream_write(invite->st, "TFR\r\n");
			invite->stage = INCOMING_BINARYMODE;
			mio_reset(invite->st->m, mt_msnftp_connect_parse, (void *)invite);
			invite->filedata = pmalloc(invite->p, j_atoi(invite->applicationfilesize, 0));
			return r_DONE;
		}
	}
	else if(j_strcmp(mt_packet_data(mp, 0), "CCL") == 0) {
		log_debug(ZONE, "Transfer cancelled");
	}
	
	return r_ERR;
}

void mt_msnftp_packets(mio m, int state, void *arg, char *buffer, int bufsz) {
	mti ti = (mti)arg;
	log_debug(ZONE, "Received from msnftp transfer %s", buffer);
}

char * mt_msnftp_grabfield(pool p, char *text, char *field) {
	// Looks through 'text' for a line beginning with 'field', then returns all the text after the colon
	// Expects lines formatted as below
/*
     Application-Name: File Transfer
     Application-GUID: {5D3E02AB-6190-11d3-BBBB-00C04F795683}
     Invitation-Command: INVITE
     Invitation-Cookie: 85366
     Application-File: Autoexec.bat
     Application-FileSize: 187
     Connectivity: N
*/

	char *str = 0;
	char *value = 0;
	int pos = 0;
	
	// Grab the field, and move to past the spaces & colons
	str = strstr(text, field);
	if(str == 0) {
		return 0; // Field not found
	}
	str += strlen(field);
	while(str[0] == ':' || str[0] == ' ') {
		str++;
	}
	
	// Look for the end of line
	pos = 0;
	while(str[pos] != '\r') {
		pos++;
	}
	
	// Allocate some memory and save the value
	value = pmalloc(p, pos + 1);
	strncpy(value, str, pos);
	value[pos] = 0;
	
	return value;
}

void mt_msnftp_checkprivate(msnftp_invite invite) {
	/* We need to create a new private switchboard if the switchboard we have now
		- is non-existant
		- is closed (maybe changed into a sbroom)
		- is empty (the person left)
		- has extra people in it
	*/
	sbchat sc = 0;
	session s = invite->s;
	if(	invite->privatesb == 0 ||
		invite->privatesb->state == sb_CLOSE ||
		invite->privatesb->users == NULL ||
		invite->privatesb->users->next != NULL
	  )
	{
		log_debug(ZONE, "Creating new private switchboard for invitations for session %s", jid_full(invite->s->id));
		sc = mt_chat_new(s,invite->remoteperson);
		sc->msnftp_invite_cookie = invite->invitationcookie;
		mt_stream_register(s->st,&mt_chat_xfr,(void *) sc);
		mt_cmd_xfr_sb(s->st);
		invite->privatesb = sc;
	}
}

msnftp_invite mt_msnftp_create(session s, char *remoteperson, char *text, sbchat sc) {
	msnftp_invite invite = 0;
	pool p = pool_new();
	
	// Check if file transfers are turned on
	if(s->ti->filetransfer == 0) return NULL;
	
	invite = pmalloc(p, sizeof(_msnftp_invite));
	invite->s = s;
	invite->p = p;
	invite->remoteperson = pstrdup(p, remoteperson);
	invite->stage = NOTHING;
	invite->invitationcookie = mt_msnftp_grabfield(p, text, "Invitation-Cookie");
	if(invite->invitationcookie == 0) {
		pool_free(p);
		return NULL;
	}
	invite->applicationguid = mt_msnftp_grabfield(p, text, "Application-GUID");
	if(strcmp(invite->applicationguid, MSNFTP_GUID) != 0) {
		pool_free(p);
		return NULL;
	}
	invite->applicationfile = mt_msnftp_grabfield(p, text, "Application-File");
	if(invite->applicationfile == 0) {
		pool_free(p);
		return NULL;
	}
	invite->applicationfilesize = mt_msnftp_grabfield(p, text, "Application-FileSize");
	if(invite->applicationfilesize == 0) {
		pool_free(p);
		return NULL;
	}
	
	// Create a private switchboard
	invite->privatesb = sc;
	mt_msnftp_checkprivate(invite);
	
	// Null out some other variables
	invite->ipaddress = 0;
	invite->ipaddressalt = 0;
	invite->port = 0;
	invite->portalt = 0;
	invite->authcookie = 0;
	invite->spoolbuff = 0;
	invite->bytesreceived = 0;
	invite->filedata = 0;
	
	xhash_put(s->msnftp_invites, invite->invitationcookie, invite);
	
	return invite;
}

msnftp_invite mt_msnftp_find(session s, char *invitationcookie) {
	msnftp_invite invite = 0;
	
	if(s->ti->filetransfer == 0) return NULL;
	if(invitationcookie == 0) return NULL;
	
	invite = (msnftp_invite) xhash_get(s->msnftp_invites, invitationcookie);
	
	return invite;
}







// FIXME, I'm pretty sure this code will only work on Linux. Implementations for
// other platforms are welcome. Either that, or send me a mail saying that this
// code works on your system and I'll add some more #ifdefs

#ifdef linux
#include <netinet/in.h>
#include <net/if.h>             //where if_nameindex() was hiding
#include <linux/sockios.h>      //for ioctl() arg
#include <sys/ioctl.h>

//thanks linus
#define NIPQUAD(addr)   ((unsigned char *)&addr)[0], \
			((unsigned char *)&addr)[1], \
			((unsigned char *)&addr)[2], \
			((unsigned char *)&addr)[3]

void mt_msnftp_fillourip(msnftp_invite invite)
{
	int sock = 0;
	struct ifreq ifreq;
	struct sockaddr_in *saptr = NULL;
	struct if_nameindex *iflist = NULL, *listsave = NULL;
	
#ifndef SIOCGIFADDR
	log_warn(ZONE, "Newer kernel needed for file transfer support. See msnftp.c:mt_msnftp_fillip()");
	return;
#endif
	//need a socket for ioctl()
	if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		log_warn(ZONE, "Socket error in msnftp.c:mt_msnftp_fillip()");
		return;
	}
	
	//returns pointer to dynamically allocated list of structs
	iflist = listsave = if_nameindex();
	
	//walk thru the array returned and query for each
	//interface's address
	for(iflist; *(char *)iflist != 0; iflist++){
		//copy in the interface name to look up address of
		strncpy(ifreq.ifr_name, iflist->if_name, IF_NAMESIZE);
		
		//get the address for this interface
		if(ioctl(sock, SIOCGIFADDR, &ifreq) != 0){
			log_warn(ZONE, "ioctl() error in msnftp.c:mt_msnftp_fillip()");
			return;
		}
		
		//save the addresses
		saptr = (struct sockaddr_in *)&ifreq.ifr_addr;
		
		// We don't want the loopback interface
		if(strstr(ifreq.ifr_name, "lo") != 0) continue;
		// Fill the addresses
		if(invite->ipaddress == 0) {
			invite->ipaddress = pmalloc(invite->p, 17); // Enough space to hold an IP address
			sprintf(invite->ipaddress, "%d.%d.%d.%d", NIPQUAD(saptr->sin_addr.s_addr));
			invite->port = pmalloc(invite->p, 10);
			strcpy(invite->port, "6891");
		}
		else if(invite->ipaddressalt == 0) {
			invite->ipaddressalt = pmalloc(invite->p, 17); // Enough space to hold an IP address
			sprintf(invite->ipaddressalt, "%d.%d.%d.%d", NIPQUAD(saptr->sin_addr.s_addr));
			invite->portalt = pmalloc(invite->p, 10);
			strcpy(invite->portalt, "11178");
			break;
		}
		
	}
	//free the dynamic memory kernel allocated for us
	if_freenameindex(listsave);
}

#else // linux
void mt_msnftp_fillourip(msnftp_invite invite) {
	log_warn(ZONE, "No implementation of mt_msnftp_fillip() for your platform. Please look in msnftp.c for details.");
}
#endif // linux


void mt_msnftp_sendpacket(msnftp_invite invite, char *text) {
	mt_msnftp_checkprivate(invite);
	
	if(invite->privatesb->state != sb_READY) {
		if(invite->spoolbuff == 0) {
			invite->spoolbuff = spool_new(invite->p);
		}
		spool_add(invite->spoolbuff, text);
	}
	else {
		mt_cmd_msg(invite->privatesb->st, "N", text);
	}
}


/*
void mt_msnftp_nextstep(msnftp_invite invite) {
	session s = invite->s;
	spool sp = spool_new(invite->p);
	
	if(s->ti->filetransfer == 0) return;
	
	switch(invite->stage) {
		case OUTGOING_INVITE:
			spool_add(sp, MIME_INVITE);
			spool_add(sp, "\r\n");
			spool_add(sp, "Application-Name: File Transfer\r\n");
			spool_add(sp, "Application-GUID: ");
			spool_add(sp, MSNFTP_GUID);
			spool_add(sp, "\r\n");
			spool_add(sp, "Invitation-Command: INVITE\r\n");
			spool_add(sp, "Invitation-Cookie: ");
			spool_add(sp, invite->invitationcookie);
			spool_add(sp, "\r\n");
			spool_add(sp, "Application-File: ");
			spool_add(sp, invite->applicationfile);
			spool_add(sp, "\r\n");
			spool_add(sp, "Application-FileSize: ");
			spool_add(sp, invite->applicationfilesize);
			spool_add(sp, "\r\n");
			mt_msnftp_sendpacket(invite, spool_print(sp));
			break;
		
		case OUTGOING_ACCEPT:
			spool_add(sp, MIME_INVITE);
			spool_add(sp, "\r\n");
			spool_add(sp, "Invitation-Command: ACCEPT\r\n");
			spool_add(sp, "Invitation-Cookie: ");
			spool_add(sp, invite->invitationcookie);
			spool_add(sp, "\r\n");
			spool_add(sp, "IP-Address: ");
			spool_add(sp, invite->ipaddress);
			spool_add(sp, "\r\n");
			spool_add(sp, "Port: ");
			spool_add(sp, invite->port);
			spool_add(sp, "\r\n");
			if(invite->ipaddressalt != 0 && invite->portalt != 0) {
				spool_add(sp, "IP-Address-Internal: ");
				spool_add(sp, invite->ipaddressalt);
				spool_add(sp, "\r\n");
				spool_add(sp, "PortX: ");
				spool_add(sp, invite->portalt);
				spool_add(sp, "\r\n");
			}
			spool_add(sp, "AuthCookie: ");
			spool_add(sp, invite->authcookie);
			spool_add(sp, "\r\n");
			spool_add(sp, "Launch-Application: FALSE\r\n");
			spool_add(sp, "Request-Data: IP-Address:\r\n");
			mt_msnftp_sendpacket(invite, spool_print(sp));
			break;
		
		case OUTGOING_CANCEL:
			spool_add(sp, MIME_INVITE);
			spool_add(sp, "\r\n");
			spool_add(sp, "Invitation-Command: CANCEL\r\n");
			spool_add(sp, "Invitation-Cookie: ");
			spool_add(sp, invite->invitationcookie);
			spool_add(sp, "\r\n");
			spool_add(sp, "Cancel-Code: ");
			if(invite->cancelcode != 0) {
				spool_add(sp, invite->cancelcode);
			}
			else {
				spool_add(sp, "REJECT");
			}
			spool_add(sp, "\r\n");
			mt_msnftp_sendpacket(invite, spool_print(sp));
			break;
		
		case INCOMING_INVITE:
		case INCOMING_ACCEPT:
		case INCOMING_CANCEL:
			break;
	}
}
*/

void mt_msnftp_offertoserve(msnftp_invite invite) {
	session s = invite->s;
	spool sp = spool_new(invite->p);
	
	invite->stage = OUTGOING_OFFERTOSERVE_SENT;
	
	if(s->ti->filetransfer == 0) return;
	
	// Fill IPs and ports
	mt_msnftp_fillourip(invite);
	
	// Generate an authcookie (that is used on out of band connect)
	invite->authcookie = pmalloc(invite->p, 15);
	sprintf(invite->authcookie, "%d", rand()%2147483647);
	
	// Construct and send the ACCEPT packet
	spool_add(sp, MIME_INVITE);
	spool_add(sp, "Invitation-Command: ACCEPT\r\n");
	spool_add(sp, "Invitation-Cookie: ");
	spool_add(sp, invite->invitationcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "IP-Address: ");
	spool_add(sp, invite->ipaddress);
	spool_add(sp, "\r\n");
	spool_add(sp, "Port: ");
	spool_add(sp, invite->port);
	spool_add(sp, "\r\n");
	if(invite->ipaddressalt != 0 && invite->portalt != 0) {
		spool_add(sp, "IP-Address-Internal: ");
		spool_add(sp, invite->ipaddressalt);
		spool_add(sp, "\r\n");
		spool_add(sp, "PortX: ");
		spool_add(sp, invite->portalt);
		spool_add(sp, "\r\n");
	}
	spool_add(sp, "AuthCookie: ");
	spool_add(sp, invite->authcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "Launch-Application: FALSE\r\n");
	spool_add(sp, "Request-Data: IP-Address:\r\n");
	spool_add(sp, "\r\n");
	mt_msnftp_sendpacket(invite, spool_print(sp));
}

void mt_msnftp_accepttransfer(msnftp_invite invite) {
	session s = invite->s;
	spool sp = spool_new(invite->p);
	
	if(s->ti->filetransfer == 0) return;
	
	// Construct and send the ACCEPT packet
	spool_add(sp, MIME_INVITE);
	spool_add(sp, "Invitation-Command: ACCEPT\r\n");
	spool_add(sp, "Invitation-Cookie: ");
	spool_add(sp, invite->invitationcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "Launch-Application: FALSE\r\n");
	spool_add(sp, "Request-Data: IP-Address:\r\n");
	spool_add(sp, "\r\n");
	mt_msnftp_sendpacket(invite, spool_print(sp));
	
	invite->stage = INCOMING_ACCEPT_SENT;
}

void mt_msnftp_starttransfer(msnftp_invite invite, char *body) {
	pool p = invite->p;
	
	log_debug(ZONE, "Inside mt_msnftp_starttransfer");
	
	invite->ipaddress = mt_msnftp_grabfield(p, body, "IP-Address");
	invite->ipaddressalt = mt_msnftp_grabfield(p, body, "IP-Address-Internal");
	invite->port = mt_msnftp_grabfield(p, body, "Port");
	invite->portalt = mt_msnftp_grabfield(p, body, "PortX");
	invite->authcookie = mt_msnftp_grabfield(p, body, "AuthCookie");
	
//	invite->st = mt_stream_connect(invite->ipaddress, j_atoi(invite->port, 6891), &mt_msnftp_connect_packets, (void *)invite);
//	invite->st = mt_stream_connect("192.168.1.2", j_atoi(invite->port, 6891), &mt_msnftp_connect_commands, (void *)invite);
//	mt_stream_write(invite->st,"VER MSNFTP\r\n");
	
	invite->stage = INCOMING_TRANSFERSTARTED;
}

void mt_msnftp_flush(session s, char *invitationcookie) {
	msnftp_invite invite = 0;
	struct spool_node *tmp;
	
	invite = mt_msnftp_find(s, invitationcookie);
	if(invite == 0) return;
	
	// Now we need to send any buffered data
	if(invite->spoolbuff == 0) return;
	invite->spoolbuff->last = 0;
	while(invite->spoolbuff->first != NULL) {
		tmp = invite->spoolbuff->first;
		invite->spoolbuff->first = tmp->next;
		mt_cmd_msg(invite->privatesb->st, "N", tmp->c);
	}
}

void mt_msnftp_sbhandle(session s, char *remoteperson, char *body, sbchat sc) {
	char *invitationcommand = 0;
	char *invitationcookie = 0;
	msnftp_invite invite = 0;
	
	// Check if file transfers are turned on
	if(s->ti->filetransfer == 0) return;
	
	// Parse the invitation
	invitationcommand = mt_msnftp_grabfield(s->p, body, "Invitation-Command");
	invitationcookie = mt_msnftp_grabfield(s->p, body, "Invitation-Cookie");
	log_debug(ZONE, "Parsing invitation packet: %s %s", invitationcommand, invitationcookie);
	
	if(strcmp(invitationcommand, "INVITE") == 0) {
		log_debug(ZONE, "Received INVITE for session %s", jid_full(s->id));
		// Create an invitation object to handle this
		invite = mt_msnftp_create(s, remoteperson, body, sc);
		invite->stage = INCOMING_INVITE_RECEIVED;
		// Send out a reply to the invitation
		log_debug(ZONE, "Sending ACCEPT for session %s", jid_full(s->id));
		mt_msnftp_accepttransfer(invite);
	}
	else if(strcmp(invitationcommand, "ACCEPT") == 0) {
		invite = mt_msnftp_find(s, invitationcookie);
		if(invite == 0) return;
		switch(invite->stage) {
			case INCOMING_ACCEPT_SENT:
				mt_msnftp_starttransfer(invite, body);
				break;
			
			default:
				log_debug(ZONE, "Some other state");
		}
	}
}

 
 
